#!/usr/local/bin/perl5 -w

# Concept test score lister.

use strict;

use CGI ':standard';
use CGI::Carp 'fatalsToBrowser';

# FMC better place?
use lib '.';
use TestBase;

# FMC factor
# Single pixel GIF trick
use constant PIXELGIF => '/icons/1p-blue.gif';

### Useful
my @letter =
  ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
   'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z');

sub answer_prefix {
  my ($num) = @_;

  if ($num == 0) {
    '(None)'
  }
  else {
    $letter[$num-1]
  }
}


# Parameters
use constant CHOOSE => 'choose';
use constant DISPLAYDATA => 'displaydata';
use constant DISPLAYACTION => 'displayaction';

use constant COURSE => 'course';
use constant TESTNAME => 'testname';

### Main

if (param(CHOOSE)) {
  my $course = param(COURSE);
  my $test_name = param(TESTNAME);

  # FMC need factor out with testgen.cgi.
  my $course_dir = course_dir($course);
  if (! -d $course_dir) {
    my $title = 'No such course';
    print
      header,
      start_html($title),
      h1($title),
      "No such course: ",
      em($course),
      ".",
      end_html;
    exit 0;
  }

  my $test_dir = test_dir($course, $test_name);
  my $stored = $test_dir . '/' .
    $test_name . TESTSTORED;
  if (! -s $stored) {
    my $title = 'No such test';
    print
      header,
      start_html($title),
      h1($title),
      "No such test for course ",
      em($course),
      ": ",
      em($test_name),
      ".",
      end_html;
    exit 0;
  }

  chdir $test_dir or
    die "$test_dir: $!\n";

  if (param(DISPLAYDATA)) {
    print_data(param(DISPLAYDATA));
  }
  else {
    print_data_choices();
  }
}
else {
  print_choose();
}
### End main

# $data is name of data file to display.
sub print_data {
  my ($data) = @_;

  open FH, $data or
    die "$data: $!\n";

  # Slurp all in.
  my @lines = <FH>;
  close FH;

  # Array of counts for each answer.
  my @data;

  # FMC analyzescores factor please

  # Depend on old/new format:
  # FMC should generate XML
  #
  # 128.x.y.z
  # first second
  # (optional blank line)

  # Sigh.  Didn't store total number of answers.
  my @largest_answer;

  # Parse.
  my $qcount = 0; # count the number of questions
  foreach (@lines) {
    next if /^\s*$/;

    if (/^([a-zA-Z]+)/) {
      # Skip student, but reset qcount
      $qcount = 0;
    }
    else {
      # a question
      my @answers = split;

      # Count last one.
      my $last_answer = $answers[$#answers];
      if ($last_answer > $largest_answer[$qcount]) {
        $largest_answer[$qcount] = $last_answer;
      }

      $data[$qcount][$last_answer]++;

      $qcount++;
    }
  }

  my $title = "Score data: $data";
  print
    header,
    start_html($title),
    h1($title);

  my $i;
  for( $i = 0; $i < scalar(@data); $i++ ) {
    my $rows = gen_rows($data[$i]);

    print 
      "Question ",($i+1),':',
      table(TR($rows)),
      p;
  }

  print 
    hr,
    pre(\@lines),

    end_html;
}

# FMC the following from conceptmonitor.cgi
sub gen_rows {
  my ($item) = @_;
  my $answer_num = 0;

  my $sum = sum($item);

  my @rows = map {
    my $count = $_ + 0;

    # Guess layout!
    my $bar = $count == 0 ?
      '' :
        do {
          my $percentage = $count * 100 / $sum;

          img({ -src => PIXELGIF,
                -alt => "$percentage%",
                -height => 20,
                -width => 4 * $percentage
              });
        };

    # Triple
    my @data = (answer_prefix($answer_num++) . '.', $count, $bar);


    td(\@data)
  } @$item;

  # Put unanswered at end.
  my $none = shift @rows;
  push(@rows, $none);

  \@rows
}


# Return sum of all elements in an array (ref).
sub sum {
  my ($item) = @_;
  my $sum = 0;
  foreach my $i (@$item) {
    $sum += $i
  }
  $sum
}


sub print_data_choices () {
  my $g = '*' . DATASUFFIX;
  my @files = glob($g);

  my $title = 'Choose Score Data to Display';
  print
    header(-pragma => 'no-cache'),
    start_html($title),
    h1($title),
    start_form,

    hidden(-name => CHOOSE),
    hidden(-name => COURSE),
    hidden(-name => TESTNAME),

    scrolling_list(-name => DISPLAYDATA,
                   -values => \@files),
    hr,
    submit(-name => DISPLAYACTION,
           -value => 'Display Score Data'),
    end_form,
    end_html;
}

# Fields:  COURSE, TESTNAME
# CHOOSE
sub print_choose () {
  my $title = 'Examine Concept Test Score Data';

  print
    header,
    start_html($title),
    h1($title),
    start_form,

    'Course code:',
    br,
    textfield(-name => COURSE,
              -size => 20),
    p,

    'Test:',
    br,
    textfield(-name => TESTNAME,
              -size => 20),
    p,

    submit(-name => CHOOSE,
           -value => 'Enter'),

    end_form,
    end_html;
}
