#!/usr/local/bin/perl5 -w

# Concept test.  Very simple.
# Allow multiple resubmissions.
# FMC currently don't check if test is a concept test or ordinary test.
# Important difference in the kind of data stored.

# For student to run.
#
# Warning: not secure.  Student could take intermediate HTML and hack
# it and cause bad things to happen upon submission of a form.

use strict;

use CGI ':standard';
use CGI::Carp 'fatalsToBrowser';

use POSIX;
use Fcntl ':flock';
use Storable;
use MLDBM qw(DB_File Storable);
#use DB_File;

# FMC better place?
use lib '.';
use TestBase;

# Parameters
use constant LOGIN => 'login';
use constant USER => 'user';
use constant TESTCONCEPT => 'testconcept';
use constant COURSE => 'course';
use constant TESTNAME => 'testname';
use constant SUBMIT => 'submit';
use constant NEWTEST => 'newtest';
use constant CURRENT_TESTID => 'current_testid';

### Main
autoEscape(0);

my @responses;

if (! cookie(-name => 'user') ) {
    # assemble redirect string

    my $redirstr = 'register.cgi?' . TESTCONCEPT . '=1&';

    if (defined param(COURSE)) {
	$redirstr .= COURSE . '=' . CGI->escape(param(COURSE)) . '&';
    }
    if (defined param(TESTNAME)) {
	$redirstr .= TESTNAME . '=' . CGI->escape(param(TESTNAME)) . '&';
    }

    # send the redirect

    print redirect($redirstr);
}
elsif (param(LOGIN)) {
  my $submit = param(SUBMIT);
  my $user = cookie(-name => 'user');
  my $course = param(COURSE);
  my $test_name = param(TESTNAME);
  my $newtest = param(NEWTEST);

  my $course_dir = course_dir($course);
  if (! -d $course_dir) {
    my $title = 'No such course';
    print
      header,
      start_html($title),
      h1($title),
      "No such course: ",
      em($course),
      ".",
      end_html;
    exit 0;
  }

  my $test_dir = test_dir($course, $test_name);
  my $stored = $test_dir . '/' .
    $test_name . TESTSTORED;
  if ((! -d $test_dir) || (! -s $stored)) {
    my $title = 'No such test';
    print
      header,
      start_html($title),
      h1($title),
      "No such test for course ",
      em($course),
      ": ",
      em($test_name),
      ".",
      end_html;
    exit 0;
  }

  # Check type.
  my $test_id;
  my $test_type;
  my $stopped;

  my $dbfile = test_dir($course, $test_name) . '/' .
    $test_name . TESTDB;
  my %db;
  {
    my $db = tie(%db, 'MLDBM',
                 $dbfile,
                 O_RDONLY)
      or die "Test score database $dbfile: $!";
    my $fd = $db->fd;
    open(DB_FH, "<&=$fd") or die "dup $fd: $!";
    flock(DB_FH, LOCK_SH);
    $stopped = $db{STOPPED()} + 0;
    $test_type = $db{TESTTYPE()};
    $test_id = $db{TESTID()};
    flock(DB_FH, LOCK_UN);
    close(DB_FH);
  }
  untie %db;

  if ($submit) {
    # Check correct test.
    if (param(CURRENT_TESTID) ne $test_id) {
      my $title = 'Test has changed';
      print
        header(-pragma => 'no-cache'),
        start_html($title),
        h1($title),
        "The test has changed. ",
        "Please hit the New Test button and submit your response again.",

        start_form,
        hidden(-name => LOGIN),
        hidden(-name => USER),
        hidden(-name => COURSE),
        hidden(-name => TESTNAME),
        hidden(-name => CURRENT_TESTID),
        submit(-name => NEWTEST,
               -value => 'New Test');
        end_form,

        end_html;
      exit 0;
    }
  }
  else {
    param(CURRENT_TESTID, $test_id);
  }

  if ($test_type ne TEST_CONCEPT) {
      my $title = 'Not a concept test';
      print
        header,
        start_html($title),
        h1($title),
        "Test ",
        em($test_name),
        " of course ",
        em($course),
        " is not a concept test.",
        end_html;
      exit 0;
  }

  my $obj = retrieve($stored);

  # Globals:  LOGIN, USER, COURSE, TESTNAME
  if ($newtest) {
    # Erase all user answers.
    my $num_items = scalar @{$obj->{items}};
    foreach my $i (1 .. $num_items) {
      Delete(item_name($i));
    }
  }
  if ($submit) {
    if ($stopped != 0) {
      my $title = 'Test no longer accepts submissions';
      print
        header,
        start_html($title),
        h1($title),
        "Test ",
        em($test_name),
        " of course ",
        em($course),
        " is no longer accepting submissions.",
        end_html;
      exit 0;
    }
    @responses = save_results($obj->{items}, $user, $course, $test_name);
  }
  print_all($submit, $obj, \@responses);
}
else {
  print_login();
}

### End main



# print_login()
#
# Fields:  USER, COURSE, TESTNAME, LOGIN
sub print_login () {
  my $title = 'Concept Test Login';

  print
    header,
    start_html($title),
    h1($title),
    start_form,

    'User: ',cookie(-name=>'user'),
    p,

    'Course code:',
    br,
    textfield(-name => COURSE,
              -size => 20),
    p,
    br,

    'Test:',
    br,
    textfield(-name => TESTNAME,
              -size => 20),
    p,
    br,

    submit(-name => LOGIN,
           -value => 'Log In'),

    end_form,
    end_html;
}

sub answer_text { my $arg = shift;
  my ($answer, $correct) = @$arg;
  $answer
}

sub answer_correct { my $arg = shift;
  my ($answer, $correct) = @$arg;
  $correct
}

sub answer_rect { my $arg = shift;
  my ($answer, $correct, $rect) = @$arg;
  $rect
}


# Return the index of the correct answer for an item.
# FMC Assumes that there is exactly one correct answer.
sub item_correct { my ($question, @answers) = @$_;
  my $count = 1;
  foreach my $answer (@answers) {
    if (answer_correct($answer)) {
      return $count;
    }
    $count++;
  }

  die "Question \"$question\" is invalid and has no correct answer.\n";
}

BEGIN {
  my $base = 'item';

  # Globals: $base
  sub item_name {
    my ($count) = @_;
    $base . $count
  }
}




# Globals:  LOGIN, USER, COURSE, TESTNAME
sub print_all {
  my ($submit, $obj, $responses) = @_;

  print
    header(-pragma => 'no-cache'),
#    start_html($obj->{title}),
#    h1($obj->{title}),
    $obj->{prologue};

  print_form($submit, $obj->{title}, $obj->{items}, $responses);

  print
    $obj->{epilogue}
#    end_html;
}


# Save results in database.
# Depends on item name encoding of answers chosen.
sub save_results {
  my ($items, $user, $course, $test_name) = @_;

  my @iota = 1 .. scalar @$items;

  # Get all the user responses.
  # Note strings are coerced to integers here.
  # Also, if an item was not answered, the result will be 0.

  my @response = map { 
      my $itemnum = $_ - 1; # be careful...item names are not zero-indexed
      my ($question,@answers) = @{$items->[$itemnum]};

      if ( ref( $question ) and
	   (defined $question->{image}) ) {

	  my $result = 0;

	  # get the coordinates of the image click

	  my $x = param( item_name($_) . "\.x" );
	  my $y = param( item_name($_) . "\.y" );
	  
	  # assume non-overlapping answer areas
	  my $i;
	  for( $i = 0; $i <= $#answers; $i++ ) {
	      my ($text,$correct,$loc) = @{$answers[$i]};
	      my ($px,$py,$pw,$ph) = split /\,/, $loc;

	      if ( ( $px < $x ) and
		   ( $x < ( $px + $pw ) ) and
		   ( $py < $y ) and
		   ( $y < ( $py + $ph ) ) ) {

		  $result = $i+1;
		  last;
	      }
	  }

	  $result;
      }
      else {
	  param(item_name($_)) + 0;
      }
  } @iota;

  my $dbfile = test_dir($course, $test_name) . '/' .
    $test_name . TESTDB;
  my %db;
  {
    my $db = tie(%db, 'MLDBM',
                 $dbfile,
                 O_RDWR)
      or die "Test score database $dbfile: $!";
    my $fd = $db->fd;

    open(DB_FH, "<&=$fd") or die "dup $fd: $!";
    flock(DB_FH, LOCK_EX);

    my $u = $db{USERS()};

    # Push the most recent response for each item.
    if (exists $u->{$user}) {
      my $items = $u->{$user};
      my $count = 0;
      foreach my $answers (@$items) {
        push(@$answers, $response[$count++]);
      }
    }
    else {
      $u->{$user} = [ map { [$_] } @response ];
    }
    # Flush back.
    $db{USERS()} = $u;

    flock(DB_FH, LOCK_UN);
  }
  untie %db;

  return @response;
}


# Print entire form.
#
# Want login data to persist.
# Globals:  LOGIN, USER, COURSE, TESTNAME
# SUBMIT, NEWTEST
#
# Hack: Put title and submit button at very top of the form, for the
# sake of visibility.
sub print_form {
  my ($submit, $title, $items, $responses) = @_;

  # hack to remove submit buttons from image
  my $isimage = (ref $items->[0]->[0]) and 
                (defined $items->[0]->[0]->{image});

  my $message = $submit ? "Resubmit Responses" : "Submit Responses";
  my $submit_button = submit(-name => SUBMIT,
                             -value => $message);
  my $newtest_button = submit(-name => NEWTEST,
                             -value => 'New Test');

  if ( $isimage ) { $submit_button = " "; $newtest_button = " "; }

  print
    start_form,

    # For visibility on Jornada.
    table({ -width => '100%' },
          TR({ -valign => 'TOP' },
              td(h2($title)),
              td({ -align => 'RIGHT' },
                 $submit_button),
              td({ -align => 'RIGHT' },
                 $newtest_button)
            )
	  );

  if ( $isimage ) {
      print p("Click in the image to submit your answer.");
      if ( $submit ) {
	  #JWN assume image questions have only one item
	  my ($text,$cor,$loc,$pfx);

	  #JWN Cover the case of an invalid answer
	  if ( $responses->[0] != 0 ) {
	      ($text,$cor,$loc) = @{$items->[0]->[$responses->[0]]};
	      $pfx = answer_prefix( $responses->[0] );
	  }
	  else {
	      $pfx = "none of the above";
	      $text = "";
	  }
       
	  if ( defined $text and $text ne "" ) {
	      $text = "($text)";
	  }

	  print p("[Last answer was $pfx. $text]");
      }
  }

  print
    # Propagate all state.
    # FMC should make automatic
    hidden(-name => LOGIN),
    hidden(-name => USER),
    hidden(-name => COURSE),
    hidden(-name => TESTNAME),
    hidden(-name => CURRENT_TESTID);

  if ( $isimage ) {
      #JWN assume image questions have only one item
      print p( image_button(-name => item_name(1),
			    -src  => $items->[0]->[0]->{image},
			    -align => 'top',
			    ),
	       hidden(-name => SUBMIT,
		      -value => "Submit Responses"));
  }
  else {
    my $stritems = gen_items($submit,$items);
    my $i = 0;
    
    my $tblstr = "";
    for( $i = 0; $i < scalar(@{$stritems}); $i+=2 ) {
      my $item1;
      my $item2 = "";

      $item1 = td({-width => '50%'},
		  put_q_in_tbl( $i+1, $stritems->[$i] ));

      if ( $i+1 < scalar(@{$stritems}) ) {
	$item2 = td({-width => '50%'},
		    put_q_in_tbl( $i+2, $stritems->[$i+1] ));
      }

      $tblstr .= TR({-valign => 'top'},
		    $item1,$item2);
    }

    print table({-width => '100%'},
		$tblstr);
  }

  print
    $submit_button,
    $newtest_button,
    end_form;
}

# each text item is displayed in a table with a question no. in
# the left field and the question items in the right field
sub put_q_in_tbl {
  my ($no,$txt) = @_;

  return table(TR({-valign => 'top'},
		  td($no . '.'),
		  td($txt)
		 )
	      );
}

# Each item is named by 1-based index into the list of items.
# Each answer has as value the 1-based index into the list of answers.
sub gen_items {
  my ($submit, $items) = @_;

  my $count = 1;

  my @result = map {
    my ($question, @answers) = @$_;

    if (ref($question)) {
	$question = $question->{'text'};
    }

    my $item_num = $count;
    my $name = item_name($count++);
    my @iota = 1 .. scalar @answers;
    
    $question .
	radio_group(
		    -name => $name,
		    -values => \@iota,
		    -labels => gen_labels($submit, $item_num, \@answers),
		    -default => '-',
		    -cols => 1
		    )
	
  
  } @$items;

  \@result
}


BEGIN {
  my @letter =
    ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
     'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z');

  sub answer_prefix {
    my ($num) = @_;

    if ($num == 0) {
      '(None)'
    }
    else {
      $letter[$num-1]
    }
  }
}

# Generate the labels mapping for the answers of a given item.
# Highlight the chosen response if any.
sub gen_labels {
  my ($submit, $item_num, $answers) = @_;

  if ($submit) {
    # Indicate that a previous choice has been made.
    my $response = param(item_name($item_num)) + 0;

    # FMC send response to monitoring.
    # There should have been login, so that information is saved.

    my $count = 1;
    my %result = map {
      my $ans = answer_text $_;
      my $highlighted = ($count == $response) ?
        '[You last chose] ' . $ans :
          $ans;

      my $c = $count++;
      ($c => answer_prefix($c) . '. ' . $highlighted)
    } @$answers;
    \%result
  }
  else {
    my $count = 1;
    my %result = map {
      my $c = $count++;
      ($c => answer_prefix($c) . '. ' . answer_text $_)
    } @$answers;
    \%result
  }
}

__END__
