#!/usr/local/bin/perl5 -wT

# For teacher to monitor the results of a test in progress.
# No authentication here to check teacher.

# FMC
# Ideally, want to be incremental.
#
# Also, want to display original test along with correct answers and
# statistics.

use strict;

# Note Push
#fmc
#use CGI::Push ':standard', ':html3';
use CGI ':standard', ':html3';
use CGI::Carp 'fatalsToBrowser';

use POSIX;
use Fcntl ':flock';
use MLDBM qw(DB_File Storable);
#use DB_File;

# FMC better place?
use lib '.';
use TestBase;

# Parameters
use constant MONITOR => 'monitor';
use constant REFRESH_ON => 'refresh_on';
use constant REFRESH_OFF => 'refresh_off';

use constant COURSE => 'course';
use constant TESTNAME => 'testname';

### Main

if (param(MONITOR)) {
  my $course = param(COURSE);
  my $test_name = param(TESTNAME);

  # Delay in second between page updates.
  # FMC param?
  my $delay = 3;


  # FMC need factor out with testgen.cgi.
  my $course_dir = course_dir($course);
  if (! -d $course_dir) {
    my $title = 'No such course';
    print
      header,
      start_html($title),
      h1($title),
      "No such course: ",
      em($course),
      ".",
      end_html;
    exit 0;
  }

  my $stored = test_dir($course, $test_name) . '/' .
    $test_name . TESTSTORED;
  if (! -s $stored) {
    my $title = 'No such test';
    print
      header,
      start_html($title),
      h1($title),
      "No such test for course ",
      em($course),
      ": ",
      em($test_name),
      ".",
      end_html;
    exit 0;
  }

  my $dbfile = test_dir($course, $test_name) . '/' .
    $test_name . TESTDB;
  my %db;

  # Periodically refresh with latest results.
  my $do_current = sub {
#    my ($q, $counter) = @_;
    local(*DB_FH);

    my %result;

    {
      my $db = tie(%db, 'MLDBM',
                   $dbfile,
                   O_RDONLY)
        or die "Test score database $dbfile: $!";
      my $fd = $db->fd;
      open(DB_FH, "<&=$fd") or die "dup $fd: $!";
      flock(DB_FH, LOCK_SH);

      # Slurp into memory and process after unlocking.
      %result = %db;

      flock(DB_FH, LOCK_UN);
      close(DB_FH);
    }
    untie %db;

    my $test_type = $result{TESTTYPE()};
  if ($test_type ne TEST_ORDINARY) {
      my $title = 'Not an ordinary test';
      print
        header,
        start_html($title),
        h1($title),
        "Test ",
        em($test_name),
        " of course ",
        em($course),
        " is not an ordinary test.",
        end_html;
      exit 0;
  }

    my $table;

    if (scalar keys %{$result{USERS()}}) {
      my @rows;
      my $count = 0;
      my $length;

      foreach my $user (sort keys %{$result{USERS()}}) {
        # FMC
        # Each user has ref to array of scores.
        my @score = @{$result{USERS()}{$user}};
        if (! $count++) {
          $length = scalar @score;
        }

        # Append total score to the row of scores.
        my $total = scalar grep { $_ != 0 } @score;
        push(@score, strong($total));

        my @data = ($user, @score);
        push(@rows, td(\@data));
      }

      my @header = ('',
                    map { strong($_) } (1 .. $length),
                    strong('Total'));
      unshift(@rows, td(\@header));
      $table = table({ -border => '' },
                     caption(strong('Results by Student')),
                     TR(\@rows))
    }

    # OK, do it.
    my $title = "Results for Course \"$course\", Test \"$test_name\"";
    my $time = localtime();

    my $button = '';
    if (param(REFRESH_OFF)) {
      print header;
      $button =
        submit(-name => REFRESH_ON,
               -value => 'Auto Refresh On')
    }
    else {
      my $s = self_url;
      print header(-refresh => "$delay; URL=$s",
                   -pragma => 'no-cache');
      $button =
        submit(-name => REFRESH_OFF,
               -value => 'Auto Refresh Off')
    }

    # No header.
#    return
    print
      start_html($title),
      h1($title),
      start_form,

      $table,

      $button,

      hidden(-name => MONITOR),
      hidden(-name => COURSE),
      hidden(-name => TESTNAME),

      end_form,
      hr,
      $time,
      end_html
    };

#fmc
#  do_push(-delay => $delay,
#          -next_page => $do_current);
    &$do_current();
}
else {
  print_monitor();
}
### End main


# Fields:  COURSE, TESTNAME
# MONITOR
sub print_monitor () {
  my $title = 'Monitor test';

  print
    header,
    start_html($title),
    h1($title),
    start_form,

    'Course code:',
    br,
    textfield(-name => COURSE,
              -size => 20),
    p,

    'Test:',
    br,
    textfield(-name => TESTNAME,
              -size => 20),
    p,

    submit(-name => MONITOR,
           -value => 'Monitor Test'),

    end_form,
    end_html;
}
