#!/usr/local/bin/perl5 -w
# FMC -T

# For teacher to upload a test and give it an ID.
# FMC for now, test in XML format.
# FMC taint problems involved in creating directories.

# FMC should use "test" or "concepttest"

use strict;

use XML::Parser;
use CGI ':standard';
use CGI::Carp 'fatalsToBrowser';

use POSIX;
use Storable;
use MLDBM qw(DB_File Storable);

use File::Basename;
# Tainted!
#use File::Path;
use File::Copy;

# FMC better place?
use lib '.';
use TestBase;


# For created directories.
# FMC too loose?
use constant PERMALL => '0777';

# Parameters
use constant SUBMIT => 'submit';
use constant COURSE => 'course';
use constant TESTNAME => 'testname';
use constant UPLOAD => 'upload';

# Validator.
use constant RXP => '/usr/local/bin/rxp';

my @tags =
  qw(test concepttest title item question answer prologue epilogue);

# For taint.
$ENV{PATH} = '/bin:/usr/bin';



### Main
if (param(SUBMIT)) {
  my $course = param(COURSE);
  # FMC need to check for valid course directory name.
  unless ($course) {
    print_not_valid("$course is an invalid course name.");
    exit 1;
  }

  my $test_type = param(TESTTYPE);
  my $test_name = param(TESTNAME);
  # FMC need check valid, e.g., no ".", "..", or "/"
  unless ($test_name) {
    print_not_valid("$test_name is an invalid test name.");
    exit 1;
  }

  my $upload = param(UPLOAD);

  process_test($course, $test_type, $test_name, $upload);
}
else {
  print_upload();
}
exit 0;
### End main


sub process_test {
  my ($course, $test_type, $test_name, $upload) = @_;

  # Validate XML.
  if (-s $upload) {
    my $all = do {
      local ($/) = undef;
      <$upload>
    };
    if (my $xml = html_convert($all)) {
      # Unfortunately, error messages will be slightly unfriendly.
      my $err = `/usr/local/bin/rxp -sVV 2>&1 <<EOF
${xml}EOF`;
      if ($?) {
        print_not_valid(pre($err));
      }
      else {
        save_test($course, $test_type, $test_name, $all, $xml);
      }
    }
    else {
      die "Invalid test file\n"
    }

  }
  else {
    print_not_found($upload);
  }
}


sub print_not_valid { my ($err) = @_;
  my $title = 'Test upload unsuccessful';
  print
    header,
    start_html($title),
    h1($title),
    $err,
    end_html;
}


# Pre:  Test is valid.
#
# Copy the test into a standard place so that both client and
# server will access it from there.
#
# Also parse and dump in reasonable format for loading.
#
# FMC note permissions of directories created
sub save_test {
  my ($course, $test_type, $test_name, $all, $xml) = @_;

  my $course_dir = course_dir($course);
  my $course_dir_msg;
  if (-d $course_dir) {
    $course_dir_msg = '';
  }
  else {
    mkdir $course_dir, PERMALL or
      die "Could not create course directory for $course: $!\n";
    $course_dir_msg = p("Created course directory $course_dir.");
    chmod 0755, $course_dir
  }

  my $test_dir = test_dir($course, $test_name);
  my $test_dir_msg;
  if (-d $test_dir) {
    $test_dir_msg = p("Overwriting existing test $test_name.");
  }
  else {
    mkdir $test_dir, PERMALL or
      die "Could not create test directory $test_name for course $course: $!\n";
    $test_dir_msg = p("Created test directory $test_name for course $course.");
    chmod 0755, $test_dir
  }

  my $copied_file = $test_dir . '/' . $test_name . TESTSUFFIX;
  open COPY, "> $copied_file" or
    die "$copied_file: $!\n";
  print COPY $all;

  print header;

  # Dump.
  my $obj = parse($xml);
  my $stored_file = $test_dir . '/' . $test_name . TESTSTORED;
  store($obj, $stored_file) or
    die "Could not store test to $stored_file\n";

  # Each item has a question, then answers, hence the number of
  # answers is the length of the item - 1.
  my @num_answers = map { $#{$_} } @{$obj->{items}};

  # Create empty score database.
  my $dbfile = test_dir($course, $test_name) . '/' .
    $test_name . TESTDB;
  my %db;
  {
    my $db = tie(%db, 'MLDBM',
                 $dbfile,
                 O_CREAT | O_TRUNC | O_RDWR)
      or die "$dbfile: $!";

    $db{TESTID()} = rand;
    $db{STOPPED()} = 1;
    $db{TESTTYPE()} = $test_type;

    # Put in numbers of answers.
    $db{NUM_ANSWERS()} = \@num_answers;

    $db{USERS()} = {};
  }
  untie %db;

  my $title = 'Test upload successful';
  print
    header,
    start_html($title),
    h1($title),
    $course_dir_msg,
    $test_dir_msg,
    "Test ",
    em($test_name),
    " was validated and accepted.",
    end_html;
}

sub print_not_found { my ($upload) = @_;
  my $title = 'File not found';
  print
    header,
    start_html($title),
    h1($title),
    "Could not read ", em($upload), ".",
    end_html;
}

sub print_upload () {
  my $title = 'Test Upload';

  print
    header,
    start_html($title),
    h1($title),
    start_multipart_form,

    'Course code:',
    textfield(-name => COURSE,
              -size => 20),
    p,

    'Desired test type:',
    radio_group(-name => TESTTYPE,
                -values => [TEST_CONCEPT, TEST_ORDINARY],
                -default => TEST_CONCEPT
               ),
    p,

    'Name to give the test:',
    textfield(-name => TESTNAME,
              -size => 20),
    p,

    'Test file (in annotated HTML format):',
    br,
    filefield(-name => UPLOAD,
              -size => 60),
    p,

    submit(-name => SUBMIT,
           -value => 'Upload Test'),
    end_form,
    end_html;
}


# Return $xml, else undef if no test found.
#
# FMC not enough checking, delayed till XML.
#
# Convert HTML as string to XML as string.
sub html_convert {
  my ($all) = @_;

  my $title_text = '';
  if (my ($title) =
      $all =~ m(<title>(.*?)</title>)xi) {
    $title_text = "<title>$title</title>"
  }

  # Escape HTML
  $all =~ s/&/&amp;/sg;

  $all =~ s/</&lt;/sg;
  $all =~ s/>/&gt;/sg;

  if (my ($begin, $test, $contents, $end) =
      $all =~
      m(
        (.*)                    # $begin
        {(test|concepttest)}    # $test
        (.*?)                   # $contents
        {/\2}
        (.*)                    # $end
       )xs) {

    # Eliminate stuff between an end tag and a begin tag.
    # (Assume to be unimportant white space markup.)
    $contents =~
      s(
        ({/[^\}]+})             # $1
        [^\{]+
       )($1)xsg;

    # Convert tags {foo} and {\foo} to <foo> and </foo>
    while ($contents =~
           s(
             {(([^\ \}]+)[^\}]*)} # $1, $2
             (.*?)              # $3
             {/\2}
            )(<$1>$3</$2>)xsg
          ) {
      # Do nothing.
    }

    <<"END"
<?xml version="1.0"?>
<!DOCTYPE $test SYSTEM "test.dtd">
<$test>
$title_text
<prologue>$begin</prologue>
$contents
<epilogue>$end</epilogue>
</$test>
END
  }
  else {
    undef
  }
}


### Begin parsing

sub string_to_bool {
  my $s = shift;
  ($s eq TRUE) ? 1 : 0
}

# Do not need to validate, because already done during submission.
#
# Returns an expat extension field {obj}.
#
# Notes:  Could have added an object to Expat object for cleanliness.
sub parse {
  my ($xml) = @_;
  my $p = new XML::Parser('ErrorContext' => 2,
                          'Handlers' => {
                                         Start => \&start,
                                         End => \&end,
                                         Char => \&char,
                                         Final => \&final
                                        });
  $p->parse($xml);
}


sub final { my $xp = shift;
  $xp->{obj}
}


sub start { my $xp = shift; my $type = shift;
  if ($type eq TEST) {
    $xp->{obj}->{items} = [];
  }
  elsif ($type eq ITEM) {
    $xp->{obj}->{curItem} = [];
    push @{$xp->{obj}->{items}}, $xp->{obj}->{curItem};
  }
  elsif ($type eq TITLE) {
    $xp->{obj}->{title} = '';
  }
  elsif ($type eq QUESTION) {
    $xp->{obj}->{curQ} = {};
    push @{$xp->{obj}->{curItem}}, $xp->{obj}->{curQ};
  }
  elsif ($type eq IMAGE) {
    $xp->{obj}->{image} = '';
  }
  elsif ($type eq TEXT) {
    $xp->{obj}->{text} = '';
  }
  elsif ($type eq ANSWER) {
    $xp->{obj}->{answer_text} = '';
    if (@_) {
      &parse_answer_params( $xp, @_ );
    }
  }
  elsif ($type eq PROLOGUE) {
    $xp->{obj}->{prologue} = '';
  }
  elsif ($type eq EPILOGUE) {
    $xp->{obj}->{epilogue} = '';
  }
}

# Parse parameters to the answer tag
sub parse_answer_params{
    my($xp,@ary) = @_;

    while( @ary ) {
      my $paramtype = shift @ary;

      if ( $paramtype eq CORRECT ) {
	$xp->{obj}->{answer_correct} = string_to_bool(shift @ary);
      }
      elsif ( $paramtype eq LOC ) {
        $xp->{obj}->{answer_rect} = shift @ary;
      }
    } 
   
    if (! defined $xp->{obj}->{answer_correct} ) {
      $xp->{obj}->{answer_correct} = 0;
    }
    if (! defined $xp->{obj}->{answer_rect} ) {
      $xp->{obj}->{answer_rect} = 0;
    }
}

# Strip white space.
# FMC desirable?
sub end { my $xp = shift; my $type = shift;
  if ($type eq TITLE) {
    $xp->{obj}->{title} =~ s/^\s+//;
    $xp->{obj}->{title} =~ s/\s+$//;
  }
  elsif ($type eq IMAGE) {
    $xp->{obj}->{image} =~ s/^\s+//;
    $xp->{obj}->{image} =~ s/\s+$//;
    $xp->{obj}->{curQ}->{image} = $xp->{obj}->{image};
  }
  elsif ($type eq TEXT) {
    $xp->{obj}->{text} =~ s/^\s+//;
    $xp->{obj}->{text} =~ s/\s+$//;
    $xp->{obj}->{curQ}->{text} = $xp->{obj}->{text};
  }
  elsif ($type eq ANSWER) {
    $xp->{obj}->{answer_text} =~ s/^\s+//;
    $xp->{obj}->{answer_text} =~ s/\s+$//;
    my $answer = [$xp->{obj}->{answer_text}, $xp->{obj}->{answer_correct}, $xp->{obj}->{answer_rect}];
    push @{$xp->{obj}->{curItem}}, $answer;
  }
  elsif ($type eq PROLOGUE) {
    $xp->{obj}->{prologue} =~ s/^\s+//;
    $xp->{obj}->{prologue} =~ s/\s+$//;
  }
  elsif ($type eq EPILOGUE) {
    $xp->{obj}->{epilogue} =~ s/^\s+//;
    $xp->{obj}->{epilogue} =~ s/\s+$//;
  }
}

sub char { my $xp = shift; my $s = shift;
  my $type = $xp->current_element;

  if ($type eq TITLE) {
    $xp->{obj}->{title} .= $s;
  }
  elsif ($type eq IMAGE) {
    $xp->{obj}->{image} .= $s;
  }
  elsif ($type eq TEXT) {
    $xp->{obj}->{text} .= $s;
  }
  elsif ($type eq ANSWER) {
    $xp->{obj}->{answer_text} .= $s;
  }
  elsif ($type eq PROLOGUE) {
    $xp->{obj}->{prologue} .= $s;
  }
  elsif ($type eq EPILOGUE) {
    $xp->{obj}->{epilogue} .= $s;
  }
}

### End parsing
